// RingBuffer.cpp: implementation of the CRingBuffer class.
//
// This source file is a supplement to the article "Digitalni zvuk". The article
// was published in the #34 issue of the "PC" computer monthly, published in
// Belgrade, Yugoslavia by "PC Press"
//
// You can freely use, modify and redistribute this file, provided that this
// header is kept unchanged
//
//                                                                 Ivica Nikolic
//                                                      ivica@galeb.etf.bg.ac.yu
//
////////////////////////////////////////////////////////////////////////////////

#include "StdInclude.h"
#include "RingBuffer.h"

CRingBuffer::CRingBuffer( UINT nBlockSize, UINT nBlocks )
{
	m_nBlockSize = nBlockSize;
	m_nBlocks = nBlocks;

	m_ppbBuffer = (BYTE**)GlobalAllocPtr( GHND, m_nBlocks*sizeof(*m_ppbBuffer) );
	for ( UINT nBlock = 0; nBlock<m_nBlocks; nBlock++ )
		m_ppbBuffer[ nBlock ] = (BYTE*)GlobalAllocPtr( GHND, m_nBlockSize );

	m_hMutexBufferGuard = CreateMutex( NULL, FALSE, NULL );

// Initially, buffer is not full -> m_hEventBufferNotFull is signalled
	m_hEventBufferNotFull = CreateEvent( NULL, TRUE, TRUE, NULL );

// Initially, buffer is empty -> m_hEventBufferNotEmpty is non-signalled
	m_hEventBufferNotEmpty = CreateEvent( NULL, TRUE, FALSE, NULL );

	m_nNextFreePos = 0;
	m_nNextToReadPos = 0;
}

CRingBuffer::~CRingBuffer()
{
	for ( UINT nBlock = 0; nBlock<m_nBlocks; nBlock++ )
		GlobalFreePtr( m_ppbBuffer[nBlock] );
	GlobalFreePtr( m_ppbBuffer );

	CloseHandle( m_hMutexBufferGuard );
	CloseHandle( m_hEventBufferNotFull );
	CloseHandle( m_hEventBufferNotEmpty );
}

void CRingBuffer::PutBlock( const BYTE* pb )
{
	HANDLE hSyncObjects[] = { m_hMutexBufferGuard, m_hEventBufferNotFull };
	WaitForMultipleObjects( 2, hSyncObjects, TRUE, INFINITE );

	if ( m_nNextFreePos==m_nNextToReadPos )
		SetEvent( m_hEventBufferNotEmpty );

	memcpy( m_ppbBuffer[m_nNextFreePos], pb, m_nBlockSize );

	m_nNextFreePos++;
	m_nNextFreePos %= m_nBlocks;

	if ( m_nNextFreePos==m_nNextToReadPos )
		ResetEvent( m_hEventBufferNotFull );

	ReleaseMutex( m_hMutexBufferGuard );
}

void CRingBuffer::GetBlock( BYTE* pb )
{
	HANDLE hSyncObjects[] = { m_hMutexBufferGuard, m_hEventBufferNotEmpty };
	WaitForMultipleObjects( 2, hSyncObjects, TRUE, INFINITE );

	if ( m_nNextFreePos==m_nNextToReadPos )
		SetEvent( m_hEventBufferNotFull );

	memcpy( pb, m_ppbBuffer[m_nNextToReadPos], m_nBlockSize );

	m_nNextToReadPos++;
	m_nNextToReadPos %= m_nBlocks;

	if ( m_nNextFreePos==m_nNextToReadPos )
		ResetEvent( m_hEventBufferNotEmpty );

	ReleaseMutex( m_hMutexBufferGuard );
}

BOOL CRingBuffer::IsFull() const
{
	BOOL bRes;
	WaitForSingleObject( m_hMutexBufferGuard, INFINITE );

	bRes = WaitForSingleObject( m_hEventBufferNotFull, 0 ) == WAIT_TIMEOUT;

	ReleaseMutex( m_hMutexBufferGuard );
	return bRes;
}

BOOL CRingBuffer::IsEmpty() const
{
	BOOL bRes;
	WaitForSingleObject( m_hMutexBufferGuard, INFINITE );

	bRes = WaitForSingleObject( m_hEventBufferNotEmpty, 0 ) == WAIT_TIMEOUT;

	ReleaseMutex( m_hMutexBufferGuard );
	return bRes;
}

UINT CRingBuffer::GetNoOfFreeBlocks() const
{
	UINT nRes;
	WaitForSingleObject( m_hMutexBufferGuard, INFINITE );

	ReleaseMutex( m_hMutexBufferGuard );
	return nRes;
}

UINT CRingBuffer::GetNoOfBlocksInBuffer() const
{
	UINT nRes;
	WaitForSingleObject( m_hMutexBufferGuard, INFINITE );

	ReleaseMutex( m_hMutexBufferGuard );
	return nRes;
}
